/**
 * description
 *
 * @author eson
 *2022年7月21日-13:48:01
 */
package com.yuandian.dataflow.controller;

import java.io.Serializable;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Random;
import java.util.concurrent.ThreadLocalRandom;

import com.alipay.sofa.jraft.Status;
import com.alipay.sofa.jraft.rpc.RpcContext;
import com.alipay.sofa.jraft.rpc.RpcProcessor;
import com.google.protobuf.Any;
import com.yuandian.dataflow.statemachine.StateFactory;
import com.yuandian.dataflow.statemachine.closure.GenericClosure;
import com.yuandian.dataflow.statemachine.operate.Operate;
import com.yuandian.dataflow.statemachine.operate.Operate.OperateType;
import com.yuandian.dataflow.statemachine.rpc.RaftResponse;
import com.yuandian.dataflow.statemachine.rpc.annotations.ProcessorRaft;
import com.yuandian.dataflow.statemachine.state.State;

import lombok.Getter;
import lombok.Setter;
import lombok.extern.slf4j.Slf4j;

/**
 * description
 *
 * @author eson
 *         2022年7月21日-13:48:01
 */

@Slf4j
@ProcessorRaft
public class PacketsProcessor implements RpcProcessor<PacketsProcessor.PacketsRequest> {

    @Setter
    @Getter
    public static class PacketsRequest implements Serializable {
        private ArrayList<Any> packets = new ArrayList<>();
    }

    public static Random rand = new Random();

    @Override
    public void handleRequest(RpcContext rpcCtx, PacketsRequest request) {
        // StateServerFactory.getStateServer().updateFsmStateAsync(s, onCompleted);
        var resp = new RaftResponse<>();
        resp.setSuccess(true);
        rpcCtx.sendResponse(resp); // 因为jraft是异步. 收到数据直接返回

        try {
            log.debug("{} handler request.packets.size(): {}", StateFactory.getServerId(), request.packets.size());
            // TODO: request.packets 入库,回填, 告警 等操作

            
            Thread.sleep(ThreadLocalRandom.current().nextLong(100, 3000));

        } catch (InterruptedException e) {
            log.info(e.toString());
        } finally { // 确保 更新 最终的任务状态给master.

            // 读状态 Closure<State> 里的 getValue<State>为 State的状态
            StateFactory.readIndexState(new GenericClosure<State>() {

                @Override
                public void run(Status status) {

                    if (!status.isOk()) {
                        log.error("失败 readIndexState {}", status);
                    }

                    // readIndexState 失败后也需要直接 更新自己状态

                    var state = this.getValue(); // 获取返回的状态
                    var ws = state.getWorkers().get(StateFactory.getServerId());
                    ws.setTaskQueueSize(ws.getTaskQueueSize() - request.packets.size()); // 减少 现有的任务数量
                    ws.setUpdateAt(Instant.now()); // 设置更新时间
 
                    Operate.CallOperate(new Operate(OperateType.PUT_WORKERSTATE, ws),
                            new GenericClosure<Operate>() {
                                @Override
                                public void run(Status status) {
                                    if (!status.isOk()) {
                                        log.error("CallOperate [{}] {}", StateFactory.getServerId(), resp);
                                    }
                                }
                            });

                }
            });
        }
        ;

    }

    @Override
    public String interest() {
        return PacketsRequest.class.getName();
    }
}
